;  "280 Zzzap / Dodgem" #2001
;  For Bally Arcade/Astrocade
;  
;  Programmed by Jay Fanton
;
;  Filename (Disassembly):       280.asm
;  Filename (for MAME emulator): 280zzap.bin
;
;  This program can be assembled using ZMAC.  Make sure that HVGLIB.H
;  is in the same directory and then type:
;      zmac -i -m -o 280.bin -x 280.lst 280.asm
;
;  Version History:
;    .01 (Sept. 13, 2011)   - First attempt at disassembly.
;    .02 (October 12, 2016) - First released to the public.
;                           - This is really only one day's work.  There is
;                             a LOT left to disassemble (and comment).
;                           - I figured it would be better to let others look
;                             at what IS disassembled rather than allow the
;                             work that has been completed to sit, unseen,
;                             on my hard drive.  
;                           - Some of the comment may be wrong, in particular,
;                             the vector comments for the car may be invalid.
 
AKEYS   EQU   $0214              ; All KEYS mask
NOLINE  EQU   96                 ; NUMBER OF DISPLAYED LINES = 96
SMLFNT  EQU   $020D              ; FoNT descriptor for SMaLl font
                                 ; in the 8K On-Board ROM          	

;  CARVECT - Player's CAR VECTor
;  ??  I THINK this is right-- but I have NOT checked it yet!
CARPVECT  EQU  $4F63              ; $4F63 Vector Block
CARPVBMR  EQU  CARPVECT           ; $4F63 Vector Magic Register    
CARPSTAT  EQU  CARPVECT+VBSTAT    ; $4F64 Vector Status
CARPTIMB  EQU  CARPVECT+VBTIMB    ; $4F65 Vector Time Base
CARPDXL   EQU  CARPVECT+VBDXL     ; $4F66 Vector Delta X Lo
CARPDXH   EQU  CARPVECT+VBDXH     ; $4F67 Vector Delta X Hi
CARPVBXL  EQU  CARPVECT+VBXL      ; $4F68 Vector X Position Lo
CARPVBXH  EQU  CARPVECT+VBXH      ; $4F69 Vector X Position Hi
CARPXCHK  EQU  CARPVECT+VBXCHK    ; $4F6A Vector X Checks
CARPDYL   EQU  CARPVECT+VBDYL     ; $4F6B Vector Delta Y Lo
CARPDYH   EQU  CARPVECT+VBDYH     ; $4F6C Vector Delta Y Hi
CARPVBYL  EQU  CARPVECT+VBYL     ;  $4F6D Vector Y Postistion Lo 
CARPVBYH  EQU  CARPVECT+VBYH      ; $4F6E Vector Y Position Hi
CARPYCHK  EQU  CARPVECT+VBYCHK    ; $4F6F Vector Y Checks

        INCLUDE "HVGLIB.H"       ; Home Video Game Library

        ORG     FIRSTC

        DB      "U"              ; User Cartridge Sentinel
        
; On-Board ROM Menu Data Structure
        DW      L2007            ; Link to Head of On-Board MENU STart
        DW      L2685            ; Address of "280 ZZZAP" menu text
        DW      L2011            ; Jump here if "280 ZZZAP" selected

L2007:  DW      MENUST           ; Link to Head of On-Board MENU STart
        DW      L26A4            ; Address of "DODGEM" menu text
        DW      L200D            ; Jump here if "DODGEM" selected

L200D:  LD      A,$01            ; Dodgem Selected (Internally as Game 1)
        JR      L2012

L2011:  XOR     A                ; 280 Zzzap Selected (Internally as Game 0)
L2012:  LD      ($4F5C),A        ; Place Game 0 or 1 into memory       

        SYSSUK  GETPAR           ;  UPI GET game PARameter from user
        DW      L2674            ;  ... Prompt Address "RACE TIME"
        DB      10000011B        ;  ... Digits = 3 (with ZERO SUPPress)
        DW      $4F63            ;  ... Parameter Address
        ;
        DI      
        LD      HL,($4F63)       ; Place Initial Race Time into register 
        LD      ($4FED),HL       ; Copy Initial Race Time into memory
        LD      SP,$4F5C         ; Place internal Game Number onto stack

        SYSTEM  INTPC            ;  UPI INTerPret with Context create
        ;
        DO      SETB             ;  UPI SET Byte
        DB      $01              ;  ... Data = 1
        DW      $4FF8            ;  ... Memory Address = 20472
        ;
        DO      FILL             ;  UPI FILL memory with data
        DW      $4F5D            ;  ... Memory Address = 20317
        DW      $0069            ;  ... Byte Count = 105
        DB      $00              ;  ... Data = 0
        ;
        DO      SETOUT           ;  UPI SET some OUTput ports
        DB      NOLINE*2         ;  ... VERBL*2 = 192 ($C0) - Display 96 (or 95??) lines        
        DB      $2A              ;  ... HORCB/4 = 42
        DB      $08              ;  ... INMODE = 8, Light Pen interrupt DISABLE
                                 ;                  and Screen interrupt ENABLE
        ;
        DO      SETB             ;  UPI SET Byte
        DB      $F8              ;  ... Data = 248
        DW      $4FB8            ;  ... Memory Address = 20408
        ;
L203A:  DO      COLSET           ;  UPI COLors SET
        DW      L2327            ;  ... Table Address = 8999
        ;
        DO      FILL             ;  UPI FILL memory with data
        DW      $4F5D            ;  ... Memory Address = 20317
        DW      $0056            ;  ... Byte Count = 86
        DB      $00              ;  ... Data = 0
        ;
        DO      FILL             ;  UPI FILL memory with data
        DW      $4000            ;  ... Memory Address = 16384
        DW      $0F00            ;  ... Byte Count = 3840
        DB      $00              ;  ... Data = 0
        ;
        DO      SETW             ;  UPI SET Word
        DW      $004F            ;  ... Data Word = 79
        DW      $4F5D            ;  ... Memory Address = 20317

        ;
        DO      FILL             ;  UPI FILL memory with data
        DW      $4FC1            ;  ... Memory Address = 20417
        DW      $0003            ;  ... Byte Count = 3
        DB      $01              ;  ... Data = 1
        ;
        DO      XINTC            ;  UPI eXit INTerpreter with Context

;  Which game is selected:
;     280 Zzzap = 0
;     Dodgem    = 1
        LD      A,($4F5C)        ; Load selected game into A
        AND     A                ; Set Zero Flag if game is 280 Zzzap 
        JR      NZ,L207E         ; If Not Zero, then game is Dodgem


;  Start to Setup 280 Zzzap Screen
;  -------------------------------

; Fill top seven lines (280 bytes) with $FF (Purple)        
        SYSSUK  FILL             ;  UPI FILL memory with data
        DW      $4000            ;  ... Memory Address = 16384
        DW      $0118            ;  ... Byte Count = 280
        DB      $FF              ;  ... Data = 255

;  Setup for Draw Car Hood (Orange Portion) ??
        LD      BC,$0134         ; 
        LD      DE,$3B38
        LD      A,$AA
        LD      H,$1B
        CALL    L2319            ; Draw Routine

;  Setup for Draw Car Hood (Orange Portion) ??
        LD      BC,$0216
        LD      DE,$3947
        LD      A,$FF
        LD      H,$0D
        CALL    L2319            ; Draw Routine

        JR      L2097            ; Skip setup for Dodgem

;  Start to Setup Dodgem Screen
;  ----------------------------

L207E:  LD      HL,$4F63         ; Place Initial Race Time into register         
        LD      ($4F9B),HL       ; Copy Initial Race Time into memory

        SYSSUK  COLSET           ;  UPI COLors SET
        DW      L232B            ;  ... Table Address = 9003
        ;
        LD      A,$A8
        LD      BC,$0104
        LD      DE,$4548
        LD      IX,$4F63
        CALL    L2224

; Draw Speedometer Box for both games
L2097:  SYSSUK  RECTAN           ;  UPI paint a RECTANgle
        DB      28               ;  ... X Posision
        DB      86               ;  ... Y Posision
        DB      108              ;  ... Width = 108
        DB      9                ;  ... Height = 9
        DB      $55              ;  ... Color = 85
        ;

;  Write speedometer numbers as a string in small (4x6) characters
;  using the Alternate FONT Set. The finished speedometer will look
;  like this:
;       0 40 80 120 160 200 240
;
        LD      IX,SMLFNT
        SYSSUK  STRDIS           ;  UPI STRing DISplay
        DB      $1E              ;  ... X = 30
        DB      $57              ;  ... Y = 87
        DB      00100100B        ;  ... Options = XOR Write, On Color = 01  
        DW      L224D            ;  ... Address of Speedometer String

        LD      A,$23
        LD      I,A
        CALL    L2334
        CALL    L2298
        CALL    L2265
L20B6:  LD      HL,$267D
        LD      BC,$0818
        OTIR    
        LD      HL,$20BE
        PUSH    HL
        SYSSUK  SENTRY           ;  UPI SENse TRansition
        DW      AKEYS            ;  ... All KEYS mask
        ;
        SYSSUK  DOIT             ;  UPI DOIT table, branch to translation handler
        DW      L265F            ;  ... DO Table Address
        ;
        IN      A,($1C)
        RRCA    
        AND     $7F
        SUB     $40
        LD      ($4FB3),A
        LD      HL,$4F60
        LD      A,(HL)
        CP      $01
        JP      Z,L22C2
        AND     A
        JR      NZ,L20EB
        LD      A,($4F5F)
        CP      $EF
        JR      C,L20E9
        LD      A,$EF
L20E9:  OUT     ($17),A
L20EB:  LD      HL,$4FBE
        LD      A,(HL)
        RLCA    
        JR      C,L213C
        LD      (HL),$FF
        INC     HL
        LD      C,(HL)
        INC     HL
        LD      B,(HL)
        INC     HL
        LD      A,B
        SUB     C
        JP      P,L2100
        NEG     
L2100:  CP      (HL)
        JR      C,L2117
        LD      A,B
        XOR     C
        RLCA    
        LD      A,B
        JR      C,L210F
        CP      C
        JR      C,L2112
L210C:  SUB     (HL)
        JR      L2113
L210F:  CP      C
        JR      C,L210C
L2112:  ADD     A,(HL)
L2113:  DEC     HL
        LD      (HL),A
        JR      L213C

L2117:  INC     HL
        DEC     (HL)
        JR      NZ,L213C
        SYSSUK  RANGED           ;  UPI RANGED random number
        DB      $08              ;  ... Cutoff = 8
        ;
        CP      $01
        JR      NZ,L2128
        LD      B,A
        SYSSUK  RANGED           ;  UPI RANGED random number
        DB      $40              ;  ... Cutoff = 64
        ;
        JR      L2131
L2128:  SYSSUK  RANGED           ;  UPI RANGED random number
        DB      $7F              ;  ... Cutoff = 127
        ;
        SUB     $40
        LD      B,A
        SYSSUK  RANGED           ;  UPI RANGED random number
        DB      $20              ;  ... Cutoff = 32
        ;
L2131:  INC     A
        LD      (HL),A
        DEC     HL
        SYSSUK  RANGED           ;  UPI RANGED random number
        DB      $08              ;  ... Cutoff = 8
        ;
        INC     A
        LD      (HL),A
        DEC     HL
        DEC     HL
        LD      (HL),B
L213C:  LD      HL,$4FB8
        LD      A,(HL)
        AND     A
        JP      M,L2159
        LD      (HL),$F8
        LD      HL,$4FC4
        LD      A,(HL)
        AND     A
        INC     A
        DAA     
        LD      (HL),A
        JR      NC,L2156
        AND     A
        INC     HL
        LD      A,(HL)
        INC     A
        DAA     
        LD      (HL),A
L2156:  CALL    L2298
L2159:  LD      A,($4F5F)
        RRCA    
        RRCA    
        AND     $3F
        LD      E,A
        RRCA    
        AND     $1F
        ADD     A,E
        LD      E,A
        AND     $03
        LD      C,A
        LD      B,$00
        LD      HL,L2249
        ADD     HL,BC
        LD      D,(HL)
        LD      A,E
        RRCA    
        RRCA    
        AND     $3F
        LD      B,A
        LD      A,$18
        SUB     B
        LD      C,A
        LD      HL,$4E90
        LD      A,B
        AND     A
        JR      Z,L2186
L2181:  LD      (HL),$00
        INC     HL
        DJNZ    L2181
L2186:  LD      (HL),D
        LD      B,C
L2188:  INC     HL
        LD      (HL),$55
        DJNZ    L2188
        LD      A,($4F5C)
        AND     A
        RET     Z

        LD      IX,$4F63
        LD      A,($4F5F)
        CP      $05
        JR      NC,L21A0
        XOR     A
        JR      L21A5

L21A0:  LD      A,($4FB3)
        NEG     
L21A5:  CALL    L220E
        LD      BC,$0300
        LD      IX,$4F71
L21AF:  LD      A,(IX+$01)
        AND     A
        JR      NZ,L21BC
        PUSH    IX
        POP     DE
        SET     0,C
        JR      L21D6

L21BC:  LD      A,($4F5F)
        RRCA    
        AND     $7F
        SUB     $30
        CALL    L2219
        LD      (IX+$09),H
        LD      (IX+$08),L
        LD      A,(IX+$0B)
        CP      $10
        JR      NC,L21D6
        SET     1,C
L21D6:  EX      DE,HL
        LD      DE,$000E
        ADD     IX,DE
        EX      DE,HL
        DJNZ    L21AF
        LD      A,C
        CP      $01
        JR      NZ,L220D
        LD      A,($4F5F)
        CP      $40
        JR      C,L220D
        PUSH    DE
        POP     IX
        SYSSUK  RANGED           ;  UPI RANGED random number
        DB      $00              ;  ... Cutoff = 0
        ;
        CP      $FA
        JR      C,L220D
        SYSSUK  RANGED           ;  UPI RANGED random number
        DB      $50              ;  ... Cutoff = 80
        ;
        ADD     A,$28
        LD      E,A
        LD      D,$02
        LD      BC,$0308
        LD      A,$28
        CALL    L2224
        SYSSUK  RANGED           ;  UPI RANGED random number
        DB      $80              ;  ... Cutoff = 128
        ;
        SUB     $40
        CALL    L220E
L220D:  RET     

L220E:  CALL    L2219
        ADD     HL,HL
        LD      (IX+$03),L
        LD      (IX+$04),H
        RET     

L2219:  LD      L,A
        LD      H,$00
        RLCA    
        JR      NC,L2221
        LD      H,$FF
L2221:  ADD     HL,HL
        ADD     HL,HL
        RET     

L2224:  PUSH    IX
        POP     HL
        LD      (HL),A
        XOR     A
        INC     HL
        PUSH    HL
        LD      (HL),A
        INC     HL
        LD      (HL),A
        INC     HL
        LD      (HL),A
        INC     HL
        LD      (HL),A
        INC     HL
        LD      (HL),A
        INC     HL
        LD      (HL),E
        INC     HL
        LD      (HL),B
        INC     HL
        LD      (HL),A
        INC     HL
        LD      (HL),A
        INC     HL
        LD      (HL),A
        INC     HL
        LD      (HL),D
        INC     HL
        LD      (HL),$01
        INC     HL
        LD      (HL),C
        POP     HL
        LD      (HL),$80
        RET     

L2249:  DB      $15
        DB      $05
        DB      $01
        DB      $00

;  String for Speedometer
;  String Use Alternate FONT Set
;  Uses SMALL CHARACTERS (4 X 6) font ??
;  This string look like this:
;  0 40 80 120 160 200 240
L224D:  DB      $B0              ; 0              
        DB      $01              ; Tab Code - Skip 1 Character
        DB      $B4              ; 4
        DB      $B0              ; 0
        DB      $01              ; Tab Code - Skip 1 Character
        DB      $B8              ; 8             
        DB      $B0              ; 0
        DB      $01              ; Tab Code - Skip 1 Character 
        DB      $B1              ; 1
        DB      $B2              ; 2
        DB      $B0              ; 0
        DB      $A0              ; Space (Small Font)
        DB      $B1              ; 1 
        DB      $B6              ; 6
        DB      $B0              ; 0
        DB      $A0              ; Space (Small Font)
        DB      $B2              ; 2 
        DB      $B0              ; 0
        DB      $B0              ; 0
        DB      $A0              ; Space (Small Font)
        DB      $B2              ; 2
        DB      $B4              ; 4
        DB      $B0              ; 0
        DB      $00              ; End of Sring Delimter

;  SENTRY RETURN Code SSEC Handler
;  Sentry, SEConds timer has counted down
L2265:  LD      A,$01
        LD      ($4F61),A
        SYSSUK  DISTIM           ;  UPI DISplay TIMe
        DB      $04              ;  ... X = 4
        DB      $5A              ;  ... Y = 90
        DB      10000100B        ;  Options = Display colon and seconds,
                                 ;            On Color = 01
        ;
        LD      HL,$4FF8
        BIT     7,(HL)
        JR      Z,L22B7
        SYSSUK  STRDIS           ;  UPI STRing DISplay
        DB      $24              ;  ... X = 36
        DB      $00              ;  ... Y = 0
        DB      $07              ;  ... Options = 7
        DW      L228C            ;  ... String Address "FINISH LINE"
        ;
        SYSTEM  EMUSIC           ;  UPI End playing MUSIC
        ;

;  Game Over.  If any key on keypad is pressed then Reset system 
L227F:  SYSSUK  SENTRY           ;  UPI SENse TRansition
        DW      AKEYS            ;  ... Keymask Address = 532
        ;
        CP      $13
        JR      Z,L228B
        CP      $15
        JR      NZ,L227F
        
L228B:  RST     $00              ;  System ReSeT

L228C:  DB      "FINISH LINE",$00  

L2298:  LD      A,$01
        LD      ($4F61),A
        LD      IX,SMLFNT
        SYSSUK  DISNUM           ;  UPI DISplay NUMber
        DB      $8C              ;  ... X = 140
        DB      $5A              ;  ... Y = 90
        DB      $04              ;  ... Options = 4
        DB      $C2              ;  ... Extended = 194
        DW      $4FC5            ;  ... Number Address = 20421
        ;
        LD      E,$98
        DEC     HL
        LD      B,$42
        SYSTEM  DISNUM           ;  UPI DISplay NUMber
        ;
        SYSSUK  RECTAN           ;  UPI paint a RECTANgle
        DB      $95              ;  ... X = 149
        DB      $5E              ;  ... Y = 94
        DB      $01              ;  ... Width = 1
        DB      $01              ;  ... Height = 1
        DB      $55              ;  ... Color = 85
        ;
L22B7:  XOR     A
        LD      ($4F61),A
        RET     

;  SENTRY RETURN Code SKYD Handler
;  Sentry, KeY is now Down
L22BC:  SYSTEM  PIZBRK           ;  UPI take a PIZza BReaK
        ;
        POP     HL
        JP      L20B6

L22C2:  LD      (HL),$02
        SYSSUK  RANGED           ;  UPI RANGED random number
        DB      4                ;  ... Choose Random Number 0-3
        ;
        SYSSUK  INDEXW           ;  UPI INDEX Word by A
        DW      L22F0            ;  ... Base Address = 8944
        ;
        EX      DE,HL
        LD      DE,$0808
        LD      C,$98
        SYSTEM  STRDIS           ;  UPI STRing DISplay
        ;
        LD      A,$20
        LD      ($4FD5),A
        LD      D,$04
        LD      HL,$2668
        LD      BC,$0317
L22E0:  LD      E,(HL)
        OUT     (C),E
        DEC     C
        INC     HL
        DJNZ    L22E0
        SYSSUK  PAWS             ;  UPI PAUSE
        DB      $0E              ;  ... Interrupts = 14
        ;
        DEC     D
        JR      NZ,$22DD
        SYSTEM  EMUSIC           ;  UPI End playing MUSIC
        ;
        RET     

; Table for INDEXW Subroutine (Valid Range 0-3) 
L22F0:  DW      L22F8            ; Address of "BANG!" String
        DW      L22FE            ; Address of "BOOM!" String
        DW      L2304            ; Address of "POW!!" String
        DW      L230A            ; Address of "ZORK!" String
L22F8:  DB      "BANG!",$00  
L22FE:  DB      "BOOM!",$00  
L2304:  DB      "POW!!",$00
L230A:  DB      "ZORK!",$00

;  SENTRY RETURN Code SCT0 Handler
;  Counter-Timer 0 has counted down
L2310:  LD      SP,$4F5C
        DI      
        SYSTEM  INTPC            ;  UPI INTerPret with Context create
        ;
        DO      MJUMP            ;  UPI Macro JUMP to interpreter subroutine
        DW      L203A            ;  ... Macro Address = 8250

L2319:  SYSTEM  RECTAN           ;  UPI paint a RECTANgle
        ;
        DEC     E
        INC     C
        INC     C
        LD      L,A
        LD      A,B
        ADD     A,D
        LD      D,A
        DEC     H
        LD      A,L
        JR      NZ,L2319
        RET     

;  Color Table
L2327:  DB      $1A
        DB      $5C
        DB      $07
        DB      $00

;  Color Table for Dodgem ??
L232B:  DB      $FC             ; color 3 Left  = 
        DB      $6A             ; color 2 Left  =
        DB      $87             ; color 1 Left  =
        DB      $35             ; color 0 Left  =
        DB      $00             ; color 3 Right = Black
        DB      $34             ; color 2 Right =
        DB      $23             ; color 1 Right =
        DB      $EA             ; color 0 Right =

L2333:  DB      $24

L2334:  EX      AF,AF'
        EXX     
        PUSH    IX
        LD      HL,($4F60)
        LD      A,H
        OR      L
        LD      HL,$4F62
        OR      (HL)
        JP      NZ,L2493
        LD      A,($4F5C)
        AND     A
        JP      Z,L23FD
        LD      IX,($4F9B)
        LD      B,$04
L2351:  BIT     6,(IX+$01)
        JR      Z,L235E
        CALL    L23EC            ; Write Car Pattern at new position ??
        RES     6,(IX+$01)
L235E:  BIT     7,(IX+$01)
        JR      Z,L23A6
        BIT     7,(IX+$00)
        JR      Z,L2379
        LD      A,($4F5F)
        RRCA    
        RRCA    
        RRCA    
        AND     $1F
        LD      L,A
        LD      A,$45
        SUB     L
        LD      (IX+$0B),A
L2379:  LD      HL,L23F9         ;  Limit Table Address
        SYSTEM  VECT             ;  UPI VECTor move coordinate pair
        ;
        CALL    L23EC            ; Write Car Pattern at new position ??
        BIT     7,(IX+$00)
        JR      Z,L239A
        IN      A,($08)
        AND     $02
        JR      NZ,L2393
        BIT     3,(IX+$07)
        JR      Z,L23A4
L2393:  LD      A,$01
        LD      ($4F60),A
        JR      L23A4

L239A:  BIT     3,(IX+$0C)
        JR      Z,L23A4
        RES     7,(IX+$01)
L23A4:  LD      B,$01
L23A6:  LD      DE,$000E
        ADD     IX,DE
        PUSH    IX
        POP     DE
        LD      A,E
        CP      $9B
        JR      NZ,L23B7
        LD      IX,$4F63
L23B7:  DJNZ    L2351
        LD      ($4F9B),IX
        LD      HL,$4004
        LD      A,($4FBB)
        LD      BC,$0A00
        CALL    L2579
        LD      A,($4FBA)
        LD      ($4FBB),A
        LD      BC,$0AAA
        CALL    L2579
        LD      HL,$4F64
        LD      DE,$000D
        LD      B,$05
L23DD:  LD      A,(HL)
        INC     HL
        RLCA    
        JR      NC,L23E3
        INC     (HL)
L23E3:  ADD     HL,DE
        DJNZ    L23DD
        CALL    L254A
        JP      L2493

L23EC:  LD      A,(IX+$0D)
        OUT     (XPAND),A        ; eXPANDer pixel definition port        
        LD      HL,L268E         ; Pattern Address (Dodgem Car Pattern)
        IN      A,(INTST)        ; read INTercept STatus ??        
        SYSTEM  VWRITR           ;  UPI Vector WRITe Relative
        ;
        RET     

;  VECT Limit Table Address 
L23F9:  DB      20               ; X Lower Limit
        DB      132              ; X Upper Limit
        DB      0                ; Y Lower Limit
        DB      76               ; Y Upper Limit

L23FD:  INC     (HL)
        LD      A,$04
        OUT     ($19),A
        LD      IX,$4FAE
        CALL    L24B7
        LD      IX,$4FA9
        CALL    L24B7
        LD      A,($4FBA)
        AND     $07
        OR      $38
        LD      C,A
        LD      ($4FBC),A
        CALL    L2592
        XOR     A
        LD      ($4FBC),A
        CALL    L254A
        LD      HL,($4FBD)
        ADD     HL,DE
        LD      ($4FBD),HL
        LD      A,($4FC0)
        AND     A
        JP      P,L2434
        CPL     
L2434:  AND     $3F
        LD      C,A
        LD      B,$00
        LD      HL,$277C
        ADD     HL,BC
        LD      A,(HL)
        LD      ($4FB4),A
        SRL     E
        JR      Z,L2493
        SRL     E
        SRL     E
        SRL     E
        LD      D,B
        LD      HL,$26AA
        ADD     HL,DE
        ADD     A,(HL)
        INC     HL
        INC     HL
        LD      B,(HL)
        LD      HL,$4FC0
        CALL    L24CE
        NEG     
        LD      C,A
        LD      A,($4FB3)
        AND     A
        JP      P,L2465
        CPL     
L2465:  AND     $3F
        LD      E,A
        LD      HL,$277C
        ADD     HL,DE
        LD      A,(HL)
        ADD     A,B
        LD      HL,$4FB3
        CALL    L24CE
        SUB     C
        LD      E,A
        RLCA    
        LD      A,$00
        SBC     A,$00
        LD      D,A
        LD      HL,($4F5D)
        ADD     HL,DE
        LD      A,H
        AND     A
        JR      NZ,L248E
        LD      A,L
        CP      $A0
        JR      NC,L248E
        LD      ($4F5D),HL
        JR      L2493

L248E:  LD      A,$01
        LD      ($4F60),A

L2493:  LD      HL,$4FC3
        LD      BC,$0601
        CALL    CTIMER
        CALL    STIMER
        LD      BC,$3246
        LD      A,($4F5C)
        AND     A
        JR      Z,$24AB
        LD      BC,$30A0
L24AB:  LD      A,B
        OUT     ($0D),A
        LD      A,C
        OUT     ($0F),A
        POP     IX
        EXX     
        EX      AF,AF'
        EI      
        RET     

L24B7:  LD      A,(IX+$00)
        AND     A
        RET     Z

        LD      D,A
        LD      E,(IX+$01)
        LD      H,(IX+$03)
        LD      L,(IX+$04)
        LD      A,(IX+$02)
        OUT     ($0C),A
        JP      $2645


L24CE:  AND     A
        LD      D,$00
        JP      P,$24DC
        CP      $E6
        LD      A,D
        JR      C,$24E4
        INC     A
        JR      $24E4
        PUSH    HL
        LD      HL,$26FC
        LD      E,A
        ADD     HL,DE
        LD      A,(HL)
        POP     HL
        BIT     7,(HL)
        RET     Z

        NEG     
        RET     

L24EA:  EX      AF,AF'
        EXX     
        PUSH    IX
        LD      HL,($4F60)
        LD      A,H
        OR      L
        JR      NZ,L2544
        LD      HL,$4F62
        LD      A,(HL)
        AND     A
        JR      Z,L2544
        LD      A,$04
        OUT     ($19),A
        XOR     A
        OUT     ($0C),A
        LD      (HL),A
        LD      C,$0E
        LD      IX,$4F63
L250A:  LD      H,(IX+$03)
        LD      L,(IX+$04)
        LD      A,(IX+$00)
        AND     A
        JR      Z,L2520
        LD      B,A
        LD      DE,$0027
L251A:  LD      (HL),D
        INC     HL
        LD      (HL),D
        ADD     HL,DE
        DJNZ    L251A
L2520:  LD      DE,$0005
        ADD     IX,DE
        DEC     C
        JP      NZ,L250A
        LD      A,($4FBA)
        LD      B,$07
        AND     B
        LD      C,A
        LD      IX,$4F63
L2534:  PUSH    BC
        CALL    L2592
        LD      BC,$0005
        ADD     IX,BC
        POP     BC
        LD      A,$08
        ADD     A,C
        LD      C,A
        DJNZ    L2534
L2544:  LD      BC,$30FC
        JP      L24AB

L254A:  LD      A,($4FE4)
        LD      HL,$4F5F
        AND     $10
        JR      Z,L255D
        LD      A,(HL)
        ADD     A,$02
        CP      $FA
        JR      NC,L2565
        JR      L2564
L255D:  LD      A,(HL)
        SUB     $04
        CP      $FB
        JR      NC,L2565
L2564:  LD      (HL),A
L2565:  LD      E,(HL)
        LD      D,$00
        LD      HL,($4FB9)
        ADD     HL,DE
        ADD     HL,DE
        ADD     HL,DE
        LD      ($4FB9),HL
        LD      HL,($4FB7)
        ADD     HL,DE
        LD      ($4FB7),HL
        RET     

L2579:  PUSH    HL
        AND     $07
        INC     A
        LD      DE,$0028
L2580:  ADD     HL,DE
        DEC     A
        JR      NZ,L2580
L2584:  LD      (HL),C
        LD      DE,$001F
        ADD     HL,DE
        LD      (HL),C
        LD      DE,$0121
        ADD     HL,DE
        DJNZ    L2584
        POP     HL
        RET     

L2592:  LD      B,$00
        LD      HL,$27BD
        ADD     HL,BC
        LD      A,($4FB4)
        ADD     A,(HL)
        LD      E,A
        JP      P,L25A8
        LD      L,B
        CP      $E6
        JR      C,L25AE
        INC     L
        JR      L25AE
L25A8:  LD      D,B
        LD      HL,$26FC
        ADD     HL,DE
        LD      L,(HL)
L25AE:  LD      H,B
        LD      A,($4FC0)
        RLCA    
        JR      NC,L25BB
        LD      A,L
        CPL     
        LD      L,A
        LD      H,$FF
        INC     HL

;  Does this routine draw the signposts for 280 Zzaap ??
L25BB:  EX      DE,HL
        LD      HL,($4F5D)
        ADD     HL,DE
        LD      ($4FB5),HL
        LD      HL,$26BC
        ADD     HL,BC
        LD      E,(HL)
        LD      A,E
        RRCA    
        AND     $7F
        LD      D,A
        RRCA    
        RRCA    
        AND     $07
        LD      C,A
        LD      HL,$269C
        ADD     HL,BC
        LD      A,E
        LD      E,(HL)
        PUSH    DE
        LD      HL,($4FB5)
        AND     A
        SBC     HL,BC
        LD      C,A
        PUSH    BC
        SBC     HL,BC
        SBC     HL,BC
        CALL    L25F4
        LD      BC,$0005
        ADD     IX,BC
        POP     BC
        POP     DE
        LD      HL,($4FB5)
        ADD     HL,BC
        ADD     HL,BC
L25F4:  LD      A,L
        CP      $99
        JR      NC,L260D
        LD      A,H
        AND     A
        JR      NZ,L260D
        LD      A,($4FBC)
        AND     A
        LD      A,L
        JR      Z,L2612
        AND     $03
        OR      $28
        LD      (IX+$02),A
        JR      L2616
L260D:  LD      (IX+$00),$00
        RET     

L2612:  AND     $03
        OR      $18
L2616:  OUT     (MAGIC),A
        LD      A,L
        LD      L,C
        ADD     HL,HL
        ADD     HL,HL
        ADD     HL,HL
        LD      B,H
        LD      C,L
        ADD     HL,HL
        ADD     HL,HL
        ADD     HL,BC
        RRCA    
        RRCA    
        AND     $3F
        LD      C,A
        LD      B,$00
        ADD     HL,BC
        LD      (IX+$00),D
        LD      (IX+$03),H
        LD      (IX+$04),L
        LD      A,($4FBC)
        AND     A
        JR      NZ,L2645
        LD      BC,$0027
L263C:  LD      (HL),E
        INC     HL
        LD      (HL),E
        ADD     HL,BC
        DEC     D
        JP      NZ,L263C
        RET     

L2645:  LD      A,D
        CP      $1F
        JR      NZ,L264C
        LD      D,$16
L264C:  LD      (IX+$01),E
        LD      BC,$0025
L2652:  LD      (HL),E
        INC     HL
        LD      (HL),E
        INC     HL
        LD      (HL),B
        INC     HL
        LD      (HL),B
        ADD     HL,BC
        DEC     D
        JP      NZ,L2652
        RET     

; DOIT Table
L265F:  
        JMPd    SCT0,L2310,$00   ; Sentry, Counter-Timer 0 has counted down
        JMPd    SKYD,L22BC,$00   ; Sentry, KeY is now Down    
        JMPd    SSEC,L2265,$00   ; Sentry, SEConds timer has counted down 
        DB      $EF             ; End DOIT Table ($EF is > $C0)

L2669:  DB      $FF
        DB      $3F
        DB      $8F
        DB      $EE
        DB      $3E
        DB      $4E
        DB      $88
        DB      $38
        DB      $48
        DB      $44
        DB      $34

L2674:  DB      "RACE TIME",$00  

        DB      $F2
        DB      $16
        DB      $00
        DB      $C8
        DB      $64
        DB      $32
        DB      $FC

;  Note: This string ends with the $00 used in the car pattern
L2685:  DB      "280 ZZZAP"  

;  Dodgem Car Pattern
L268E:  DB      $00,$00          ; 
        DB      $01,$0A
        DB      $5A              ; .1.11.1.
        DB      $7E              ; .111111.
        DB      $5A              ; .1.11.1.
        DB      $18              ; ...11...
        DB      $3C              ; ..1111..
        DB      $7E              ; .111111.
        DB      $18              ; ...11...
        DB      $DB              ; 11.11.11
        DB      $FF              ; 11111111
        DB      $DB              ; 11.11.11

L269C:  DB      $80
        DB      $C0
        DB      $E0
        DB      $F0
        DB      $F8
        DB      $FC
        DB      $FE
        DB      $FF

; Note: This string ends with the $00 used (in what looks like)
; the start of a table.
L26A4:  DB      "DODGEM"

L26AA:  DB      $00  
        DB      $00
        DB      $09
        DB      $0F
        DB      $13
        DB      $16
        DB      $18
        DB      $19
        DB      $1A
        DB      $1C
        DB      $1D
        DB      $1E
        DB      $20
        DB      $21
        DB      $22
        DB      $23
        DB      $24
        DB      $24
        DB      $07
        DB      $07
        DB      $07
        DB      $07
        DB      $08
        DB      $08
        DB      $08
        DB      $08
        DB      $08
        DB      $09
        DB      $09
        DB      $09
        DB      $09
        DB      $09
        DB      $09
        DB      $09
        DB      $09
        DB      $09
        DB      $0A
        DB      $0A
        DB      $0A
        DB      $0A
        DB      $0A
        DB      $0A
        DB      $0A
        DB      $0B
        DB      $0B
        DB      $0B
        DB      $0C
        DB      $0C
        DB      $0C
        DB      $0C
        DB      $0D
        DB      $0D
        DB      $0D
        DB      $0E
        DB      $0E
        DB      $0E
        DB      $0F
        DB      $0F
        DB      $10
        DB      $11
        DB      $11
        DB      $12
        DB      $13
        DB      $13
        DB      $14
        DB      $15
        DB      $15
        DB      $16
        DB      $17
        DB      $18
        DB      $1A
        DB      $1B
        DB      $1D
        DB      $1F
        DB      $21
        DB      $24
        DB      $27
        DB      $2B
        DB      $2F
        DB      $35
        DB      $3A
        DB      $3F
        DB      $01
        DB      $01
        DB      $01
        DB      $01
        DB      $01
        DB      $01
        DB      $01
        DB      $01
        DB      $01
        DB      $01
        DB      $01
        DB      $01
        DB      $02
        DB      $02
        DB      $02
        DB      $02
        DB      $02
        DB      $02
        DB      $02
        DB      $02
        DB      $02
        DB      $02
        DB      $02
        DB      $02
        DB      $02
        DB      $02
        DB      $02
        DB      $03
        DB      $03
        DB      $03
        DB      $03
        DB      $03
        DB      $03
        DB      $03
        DB      $03
        DB      $03
        DB      $03
        DB      $03
        DB      $04
        DB      $04
        DB      $04
        DB      $04
        DB      $04
        DB      $04
        DB      $04
        DB      $05
        DB      $05
        DB      $05
        DB      $05
        DB      $05
        DB      $05
        DB      $06
        DB      $06
        DB      $06
        DB      $06
        DB      $06
        DB      $07
        DB      $07
        DB      $07
        DB      $07
        DB      $07
        DB      $08
        DB      $08
        DB      $08
        DB      $08
        DB      $09
        DB      $09
        DB      $09
        DB      $0A
        DB      $0A
        DB      $0A
        DB      $0B
        DB      $0B
        DB      $0B
        DB      $0C
        DB      $0C
        DB      $0D
        DB      $0D
        DB      $0D
        DB      $0E
        DB      $0E
        DB      $0F
        DB      $0F
        DB      $10
        DB      $10
        DB      $11
        DB      $11
        DB      $12
        DB      $13
        DB      $13
        DB      $14
        DB      $15
        DB      $15
        DB      $16
        DB      $17
        DB      $17
        DB      $18
        DB      $19
        DB      $1A
        DB      $1B
        DB      $1C
        DB      $1D
        DB      $1F
        DB      $1F
        DB      $20
        DB      $21
        DB      $22
        DB      $23
        DB      $24
        DB      $25
        DB      $26
        DB      $28
        DB      $29
        DB      $2A
        DB      $2C
        DB      $2D
        DB      $2F
        DB      $30
        DB      $32
        DB      $34
        DB      $35
        DB      $37
        DB      $39
        DB      $3B
        DB      $3D
        DB      $3F
        DB      $41
        DB      $00
        DB      $81
        DB      $82
        DB      $97
        DB      $A3
        DB      $AC
        DB      $B3
        DB      $B8
        DB      $BD
        DB      $C1
        DB      $C4
        DB      $C8
        DB      $CA
        DB      $CD
        DB      $D0
        DB      $D2
        DB      $D4
        DB      $D6
        DB      $D8
        DB      $D9
        DB      $DB
        DB      $DD
        DB      $DE
        DB      $E0
        DB      $E1
        DB      $E2
        DB      $E3
        DB      $E5
        DB      $E6
        DB      $E7
        DB      $E8
        DB      $E9
        DB      $EA
        DB      $EB
        DB      $EC
        DB      $ED
        DB      $EE
        DB      $EE
        DB      $EF
        DB      $F0
        DB      $F1
        DB      $F2
        DB      $F2
        DB      $F3
        DB      $F4
        DB      $F5
        DB      $F5
        DB      $F6
        DB      $F7
        DB      $F7
        DB      $F8
        DB      $F8
        DB      $F9
        DB      $FA
        DB      $FA
        DB      $FB
        DB      $FB
        DB      $FC
        DB      $FC
        DB      $FD
        DB      $FD
        DB      $FE
        DB      $FE
        DB      $FF
        DB      $FF
        DB      $FF
        DB      $7F
        DB      $7E
        DB      $7E
        DB      $7D
        DB      $7D
        DB      $7C
        DB      $7C
        DB      $7B
        DB      $7B
        DB      $7A
        DB      $79
        DB      $79
        DB      $78
        DB      $78
        DB      $77
        DB      $76
        DB      $76
        DB      $75
        DB      $75
        DB      $74
        DB      $73
        DB      $73
        DB      $72
        DB      $71
        DB      $70
        DB      $6F
        DB      $6F
        DB      $6E
        DB      $6D
        DB      $6C
        DB      $6C
        DB      $6B
        DB      $6A
        DB      $69
        DB      $68
        DB      $68
        DB      $67
        DB      $66
        DB      $65
        DB      $64
        DB      $63
        DB      $62
        DB      $61
        DB      $60
        DB      $5F
        DB      $5D
        DB      $5C
        DB      $5B
        DB      $59
        DB      $58
        DB      $57
        DB      $55
        DB      $53
        DB      $52
        DB      $50
        DB      $4B
        DB      $4A
        DB      $49
        DB      $47
        DB      $44
        DB      $41
        DB      $3E
        DB      $3A
        DB      $35
        DB      $C5
        DB      $FF
        DB      $FF

End     ; 

Instructions from Astrocade Manual
----------------------------------

280 ZZZAP/DODGEM

     Two different high-speed car races demand quick response and raw courage. 
Enter your racing time into the computer floor the accelerator-- and go for 
speed and distance. 280 Zzzap challenges you to hold a road filled with 
dangerous hairpin turns! Dodgem puts you up against other cars driven by the 
computer. The sharper your driving skills, the more mileage you'll get!

VIDEOCADE #2001
280 ZZZAP DODGEM
Programmed by Jay Fenton

280 ZZZAP (1 player)

     Feeling the pressure of speed and time, you're off-- along the course of a 
cross country road race. Negotiate the curves as carefully and quickly as 
possible to score the most points.

STARTING THE GAME

     Choose 280 Zzzap by pressing 1 on the keypad or by using remote game 
selection (pull the trigger on hand control #1, turn the knob until 1 appears 
on the screen, then pull the trigger again).

     Enter the game time-- up to ten minutes in length.

     Enter     9   Play 9 seconds.
     Enter   9 9   Play 99 seconds.
     Enter 9 9 9   Play 9 minutes and 99 seconds.

     If the race time being entered on the keypad is 99 seconds or less, you 
must press the = sign to start the game.

PLAYING THE GAME

     Since this is a one-player game, you will need to use hand control #1. The 
hand control functions for 280 Zzzap are:

     Trigger      Controls the speed of the car-- accelerator.
     Knob         Steers the car-- steering wheel.
     Joystick     Has no function.

     The object of the game is to accumulate as much mileage as possible by 
traveling as far and as fast as you can without crashing into the white road 
poles. Every time the car crashes into the poles, you lose valuable speed and 
distance.

     The numbers in the lower center of the screen represent the speedometer. 
As the car accelerates, the speed is indicated by the bar underneath the 
numbers.

     The numbers on the right side of the speedometer record the miles 
traveled. On the left side of the speedometer is the race time remaining.

SCORING

     To compete against two or more players, enter the same race time for each 
player. The winner is the one with the highest mileage when the race has ended.

PLAYING AGAIN

     To play 280 Zzzap again, press the RESET button and follow the 
instructions for starting the game.

DODGEM (1 player)

     Test your skill as a race car driver. Although this game is similar to 280 
Zzzap, you have an additional hazard of crashing into the other cars in the 
race!

STARTING THE GAME

     Choose DODGEM by pressing 2 on the keypad or by using remote game 
selection (pull the trigger on hand control #1, turn the knob until 2 appears 
on the screen, then pull the trigger again). Enter the game time--up to ten 
minutes in length.

     Enter     9   Play 9 seconds.
     Enter   9 9   Play 99 seconds.
     Enter 9 9 9   Play 9 minutes and 99 seconds.

     If you enter a race time that is 99 seconds or less, press the = sign to 
start the game.

PLAYING THE GAME

The hand control functions for Dodgem are:

     Trigger      Controls the speed of the car-- accelerator.
     Knob         Steers the car-- steering wheel.
     Joystick     Has no function.

     The object of Dodgem is to go as far as possible without crashing into the 
race track walls, other cars in the race, or having them crash into you.

     The light colored car is the only one that you can control. The other cars 
are controlled by the Arcade.

     The numbers in the lower center of the screen represent the speedometer. 
As the car accelerates, the speed is indicated by the bar underneath the 
numbers.

     The numbers on the right side of the speedometer are the miles traveled. 
The left side shows the race time remaining.

SCORING

     A player's score is determined by the miles driven in the specified racing 
time. If two or more players are competing against each other, the same race 
time must be entered for each player.

PLAYING AGAIN

     To play Dodgem again, press the RESET button and follow the instructions 
for starting the game.
